@@ -1,3 +1,5 @@
+"""Patch asyncio to allow nested event loops."""
+
 import asyncio
 import asyncio.events as events
 import os
@@ -9,29 +11,20 @@ from heapq import heappop
 
 def apply(loop=None):
     """Patch asyncio to make its event loop reentrant."""
-    loop = loop or asyncio.get_event_loop()
-    if not isinstance(loop, asyncio.BaseEventLoop):
-        raise ValueError('Can\'t patch loop of type %s' % type(loop))
-    if getattr(loop, '_nest_patched', None):
-        # already patched
-        return
     _patch_asyncio()
-    _patch_loop(loop)
+    _patch_policy()
     _patch_task()
     _patch_tornado()
 
+    loop = loop or asyncio.get_event_loop()
+    _patch_loop(loop)
+
 
 def _patch_asyncio():
-    """
-    Patch asyncio module to use pure Python tasks and futures,
-    use module level _current_tasks, all_tasks and patch run method.
-    """
+    """Patch asyncio module to use pure Python tasks and futures."""
+
     def run(main, *, debug=False):
-        loop = events._get_running_loop()
-        if not loop:
-            loop = events.new_event_loop()
-            events.set_event_loop(loop)
-            _patch_loop(loop)
+        loop = asyncio.get_event_loop()
         loop.set_debug(debug)
         task = asyncio.ensure_future(main)
         try:
@@ -42,6 +35,15 @@ def _patch_asyncio():
                 with suppress(asyncio.CancelledError):
                     loop.run_until_complete(task)
 
+    def _get_event_loop(stacklevel=3):
+        loop = events._get_running_loop()
+        if loop is None:
+            loop = events.get_event_loop_policy().get_event_loop()
+        return loop
+
+    # Use module level _current_tasks, all_tasks and patch run method.
+    if hasattr(asyncio, '_nest_patched'):
+        return
     if sys.version_info >= (3, 6, 0):
         asyncio.Task = asyncio.tasks._CTask = asyncio.tasks.Task = \
             asyncio.tasks._PyTask
@@ -50,9 +52,25 @@ def _patch_asyncio():
     if sys.version_info < (3, 7, 0):
         asyncio.tasks._current_tasks = asyncio.tasks.Task._current_tasks
         asyncio.all_tasks = asyncio.tasks.Task.all_tasks
-    if not hasattr(asyncio, '_run_orig'):
-        asyncio._run_orig = getattr(asyncio, 'run', None)
-        asyncio.run = run
+    if sys.version_info >= (3, 9, 0):
+        events._get_event_loop = events.get_event_loop = \
+            asyncio.get_event_loop = _get_event_loop
+    asyncio.run = run
+    asyncio._nest_patched = True
+
+
+def _patch_policy():
+    """Patch the policy to always return a patched loop."""
+
+    def get_event_loop(self):
+        if self._local._loop is None:
+            loop = self.new_event_loop()
+            _patch_loop(loop)
+            self.set_event_loop(loop)
+        return self._local._loop
+
+    policy = events.get_event_loop_policy()
+    policy.__class__.get_event_loop = get_event_loop
 
 
 def _patch_loop(loop):
@@ -160,23 +178,40 @@ def _patch_loop(loop):
         """Do not throw exception if loop is already running."""
         pass
 
+    if hasattr(loop, '_nest_patched'):
+        return
+    if not isinstance(loop, asyncio.BaseEventLoop):
+        raise ValueError('Can\'t patch loop of type %s' % type(loop))
     cls = loop.__class__
     cls.run_forever = run_forever
     cls.run_until_complete = run_until_complete
     cls._run_once = _run_once
     cls._check_running = _check_running
     cls._check_runnung = _check_running  # typo in Python 3.7 source
-    cls._nest_patched = True
-    cls._num_runs_pending = 0
+    cls._num_runs_pending = 1 if loop.is_running() else 0
     cls._is_proactorloop = (
-            os.name == 'nt' and issubclass(cls, asyncio.ProactorEventLoop))
+        os.name == 'nt' and issubclass(cls, asyncio.ProactorEventLoop))
     if sys.version_info < (3, 7, 0):
         cls._set_coroutine_origin_tracking = cls._set_coroutine_wrapper
+    cls._nest_patched = True
 
 
 def _patch_task():
     """Patch the Task's step and enter/leave methods to make it reentrant."""
+
+    def step(task, exc=None):
+        curr_task = curr_tasks.get(task._loop)
+        try:
+            step_orig(task, exc)
+        finally:
+            if curr_task is None:
+                curr_tasks.pop(task._loop, None)
+            else:
+                curr_tasks[task._loop] = curr_task
+
     Task = asyncio.Task
+    if hasattr(Task, '_nest_patched'):
+        return
     if sys.version_info >= (3, 7, 0):
 
         def enter_task(loop, task):
@@ -188,30 +223,22 @@ def _patch_task():
         asyncio.tasks._enter_task = enter_task
         asyncio.tasks._leave_task = leave_task
         curr_tasks = asyncio.tasks._current_tasks
+        step_orig = Task._Task__step
+        Task._Task__step = step
     else:
-
-        def step(task, exc=None):
-            curr_task = curr_tasks.get(task._loop)
-            try:
-                step_orig(task, exc)
-            finally:
-                if curr_task is None:
-                    curr_tasks.pop(task._loop, None)
-                else:
-                    curr_tasks[task._loop] = curr_task
-
         curr_tasks = Task._current_tasks
         step_orig = Task._step
         Task._step = step
+    Task._nest_patched = True
 
 
 def _patch_tornado():
-        """
-        If tornado is imported before nest_asyncio, make tornado aware of
-        the pure-Python asyncio Future.
-        """
-        if 'tornado' in sys.modules:
-            import tornado.concurrent as tc
-            tc.Future = asyncio.Future
-            if asyncio.Future not in tc.FUTURES:
-                tc.FUTURES += (asyncio.Future,)
+    """
+    If tornado is imported before nest_asyncio, make tornado aware of
+    the pure-Python asyncio Future.
+    """
+    if 'tornado' in sys.modules:
+        import tornado.concurrent as tc  # type: ignore
+        tc.Future = asyncio.Future
+        if asyncio.Future not in tc.FUTURES:
+            tc.FUTURES += (asyncio.Future,)
